/* This file is part of the Kiretu, a Kinect reconstruction tutor.
 * 
 * http://pille.iwr.uni-heidelberg.de/~kinect01/
 * 
 * The code ist licensed to you under the terms of the GNU General Public 
 * License, version 2.0. See the GPL2 file for the text of the license or the 
 * following URL:
 * 
 * http://www.gnu.org/licenses/gpl-2.0.txt
 * 
 * If you redistribute this file in source form, modified or unmodified, you
 * have to leave this header intact and distribute it under the same terms with 
 * the GPL2 file.
 * 
 * Binary distributions must follow the binary distribution requirements of
 * the License.
 * 
 */

/** \file CloudWriter.cpp
 * 
 * Implementation of the CloudWriter-class. See header-file for description.
 * 
 * \author		Daniel Wunderlich (d.wunderlich@stud.uni-heidelberg.de)
 * \version		0.5
 * \date		2011-01-08
 */

#include "CloudWriter.h"

CloudWriter::CloudWriter(const std::string filenamePrefix) :
	filenamePrefix(filenamePrefix),
	cloud(640*480, std::vector<float>(3)),
	cloudRgb(640*480, std::vector<int>(3)),
	cloudRgbMapping(640*480, std::vector<int>(2)),
	//~ normals(640*480, std::vector<float>(3)),
	validValues(640*480),
	reconstructionSteps(5),
	printUndefined(false),
	reconstructionStepsAdded(false),
	mirrorX(1),
	mirrorY(1),
	mirrorZ(1) {
		
	undefinedColor[0] = 255;
	undefinedColor[1] = 0;
	undefinedColor[2] = 0;
}

CloudWriter::~CloudWriter() {
}

void CloudWriter::addCloud(std::vector< std::vector<float> > cloud) {
	
	this->cloud = cloud;
}

void CloudWriter::addCloudRgb(std::vector< std::vector<int> > cloudRgb) {
	
	this->cloudRgb = cloudRgb;
}

void CloudWriter::addCloudRgbMapping(
	std::vector< std::vector<int> > cloudRgbMapping) {
	
	this->cloudRgbMapping = cloudRgbMapping;
}

//~ void CloudWriter::addNormals(
	//~ std::vector< std::vector<float> > normals) {
	//~ 
	//~ this->normals = normals;
//~ }


void CloudWriter::addValidValues (std::vector<bool> validValues) {
	this->validValues = validValues;
}


void CloudWriter::addReconstructionSteps(std::vector<bool> reconstructionSteps) {
	
	this->reconstructionSteps = reconstructionSteps;
	reconstructionStepsAdded = true;
}

void CloudWriter::setUndefined(bool printUndefined) {
	this->printUndefined = printUndefined;
}

void CloudWriter::setUndefinedColor(int color[3]) {
	undefinedColor[0] = color[0];
	undefinedColor[1] = color[1];
	undefinedColor[2] = color[2];
}

void CloudWriter::setMirror(CloudMirror mirror) {
	
	switch(mirror) {
		
		case MIRROR_X:
			mirrorX = -1 * mirrorX;
			break;
		
		case MIRROR_Y:
			mirrorY = -1 * mirrorY;
			break;
		
		case MIRROR_Z:
			mirrorZ = -1 * mirrorZ;
			break;
	}
}

void CloudWriter::writeCloud() {
	
	char* filenameChar;
	
	createOutput();
	createFilename();
	
	filenameChar = new char[filename.size() + 1];
	strcpy(filenameChar, filename.c_str());
	
	std::ofstream f;
	
	f.open(filenameChar, std::ios::out);
	
	f << "ply" << std::endl;
	f << "format ascii 1.0" << std::endl;
	f << "element vertex " << cloudOutput.size() << std::endl;
	f << "property float x" << std::endl;
	f << "property float y" << std::endl;
	f << "property float z" << std::endl;
	f << "property uchar red" << std::endl;
	f << "property uchar green" << std::endl;
	f << "property uchar blue" << std::endl;
	
	//~ if (reconstructionStepsAdded && reconstructionSteps[4]) {
		//~ 
		//~ f << "property float nx" << std::endl;
		//~ f << "property float ny" << std::endl;
		//~ f << "property float nz" << std::endl;
	//~ }
	
	f << "end_header" << std::endl;
	
	for (int i = 0; i < cloudOutput.size(); i++) {
		
		// Points
		f << cloudOutput[i][0] << " ";
		f << cloudOutput[i][1] << " ";
		f << cloudOutput[i][2] << " ";
		
		// Colors
		f << (int) cloudOutput[i][3] << " ";
		f << (int) cloudOutput[i][4] << " ";
		f << (int) cloudOutput[i][5] << " ";
		
		// Normals
		//~ if (reconstructionStepsAdded && reconstructionSteps[4]) {
		//~ 
			//~ f << cloudOutput[i][6] << " ";
			//~ f << cloudOutput[i][7] << " ";
			//~ f << cloudOutput[i][8] << " ";
		//~ }
		
		f << std::endl;
	}
	
	f.close();
	
	delete[] filenameChar;
}


void CloudWriter::createOutput() {
	
	std::vector<float> vec(9);
	
	cloudOutput.clear();	// 
	
	for (int i = 0; i < cloud.size(); i++) {
		
		if (validValues[i]) {
			
			if (cloudRgbMapping[i][0] >= 0 && cloudRgbMapping[i][0] < 640 && 
				cloudRgbMapping[i][1] >= 0 && cloudRgbMapping[i][1] < 480) {
				
				vec[0] = mirrorX * cloud[i][0];
				vec[1] = mirrorY * cloud[i][1];
				vec[2] = mirrorZ * cloud[i][2];
				vec[3] = cloudRgb[(int) (cloudRgbMapping[i][1]*640 + cloudRgbMapping[i][0])][0];
				vec[4] = cloudRgb[(int) (cloudRgbMapping[i][1]*640 + cloudRgbMapping[i][0])][1];
				vec[5] = cloudRgb[(int) (cloudRgbMapping[i][1]*640 + cloudRgbMapping[i][0])][2];
				//~ vec[6] = normals[i][0];
				//~ vec[7] = normals[i][1];
				//~ vec[8] = normals[i][2];
				
				cloudOutput.push_back(vec);
				
			} else {
				
				if (printUndefined) {
										
					vec[0] = mirrorX * cloud[i][0];
					vec[1] = mirrorY * cloud[i][1];
					vec[2] = mirrorZ * cloud[i][2];
					vec[3] = undefinedColor[0];
					vec[4] = undefinedColor[1];
					vec[5] = undefinedColor[2];
					//~ vec[6] = normals[i][0];
					//~ vec[7] = normals[i][1];
					//~ vec[8] = normals[i][2];
					
					cloudOutput.push_back(vec);
				}
			}
		}
	}
}

void CloudWriter::createFilename() {
	
	time_t timeStamp;
	struct tm *timeStruct;
	char timeChar[20];
	
	filename = filenamePrefix;
	
	time(&timeStamp);
	timeStruct = localtime(&timeStamp);
	strftime(timeChar, 20, "%Y-%m-%d-%H-%M-%S", timeStruct);
	std::string timeStr(timeChar);
	
	filename += "-";
	filename += timeStr;
	
	if (reconstructionStepsAdded) {
		
		if (reconstructionSteps[0]) {
			filename += "-M";
		} else {
			filename += "-m";
		}
		
		if (reconstructionSteps[1]) {
			filename += "-D";
		} else {
			filename += "-d";
		}
		
		if (reconstructionSteps[2]) {
			filename += "-E";
		} else {
			filename += "-e";
		}
		
		if (reconstructionSteps[3]) {
			filename += "-R";
		} else {
			filename += "-r";
		}
		
		//~ if (reconstructionSteps[4]) {
			//~ filename += "-N";
		//~ } else {
			//~ filename += "-n";
		//~ }
	}
	
	filename += ".ply";
}
