/* This file is part of the Kiretu, a Kinect reconstruction tutor.
 * 
 * http://pille.iwr.uni-heidelberg.de/~kinect01/
 * 
 * The code ist licensed to you under the terms of the GNU General Public 
 * License, version 2.0. See the GPL2 file for the text of the license or the 
 * following URL:
 * 
 * http://www.gnu.org/licenses/gpl-2.0.txt
 * 
 * If you redistribute this file in source form, modified or unmodified, you
 * have to leave this header intact and distribute it under the same terms with 
 * the GPL2 file.
 * 
 * Binary distributions must follow the binary distribution requirements of
 * the License.
 * 
 */
 
/** \file FrameGrabber.cpp
 * 
 * Implementation of the FrameGrabber-class. See header-file for description.
 * 
 * \author		Daniel Wunderlich (d.wunderlich@stud.uni-heidelberg.de)
 * \version		0.5
 * \date		2011-01-08
 */

#include "FrameGrabber.h"

FrameGrabber::FrameGrabber(unsigned int numberFrames, unsigned int sdThreshold) :
	numberFrames(numberFrames),
	sdThreshold(sdThreshold),
	countFrames(0),
	validCountTotal(0)
	 {		
		frames = new int[numberFrames][640*480];
		
		for (int i = 0; i < 640*480; i++) {
			zValues[i] = 0;
			validValues[i] = false;
		}
}

FrameGrabber::~FrameGrabber() {
	
	delete[] frames;
}

bool FrameGrabber::addFrame(uint16_t* depth) {
		
	if (countFrames < numberFrames) {
		int zvalue = 0;
		
		for (int i = 0; i < 640*480; i++) {
			
			zvalue = depth[i];
						
			if (zvalue < 2047) {
				frames[countFrames][i] = zvalue;
			} else {
				frames[countFrames][i] = 0;
			}
		}
		countFrames++;
		
		return true;
	} else {
		return false;
	}
}

void FrameGrabber::addRgb(uint8_t* rgb) {
	
	for (int i = 0; i < 640*480; i++) {
		
		rgbValues[i*3 + 0] = rgb[i*3 + 0];
		rgbValues[i*3 + 1] = rgb[i*3 + 1];
		rgbValues[i*3 + 2] = rgb[i*3 + 2];
	}
}

void FrameGrabber::computeZValues() {
		
	if (countFrames == numberFrames) {
		
		int validCount = 0;		// Number of valid (zvalue < 2047) frames per pixel
		int zValue = 0;
		float sd;				// Standard deviation
		
		std::vector<float> sdVec;
		
		for (int i = 0; i < 640*480; i++) {
			
			validCount = 0;
			zValue = 0;
			sd = 0;
			
			for (int j = 0; j < numberFrames; j++) {
				if (frames[j][i] > 0) {
					zValue += frames[j][i];
					validCount++;
				}
			}
			
			if (validCount > 0) {
				zValue = zValue / validCount;
			}
			
			if (validCount > 1) {
				sd = computeSD(i, zValue, validCount);
				sdVec.push_back(sd);
			}
			
			if (validCount > 0 && sd <= sdThreshold) {
				zValues[i] = zValue;
				
				validValues[i] = true;
				validCountTotal++;
			}		
		}
				
		countFrames = 0;		
	}
}


float FrameGrabber::computeSD(int index, int zMean, int validCount) {
		
		float sd = 0;
		float z = (float) zMean;
		float frame;
		float valid = validCount;
		
		for (int j = 0; j < numberFrames; j++) {
			
			frame = frames[j][index];
			
			if (frame > 0) {
				
				sd += pow(z - frame, 2.0);
			}
		}
		
		sd = 1 / (valid - 1) * sd;
		sd = sqrt(sd);
		
		return sd;
}


int* FrameGrabber::getZValues() {
	return zValues;
}

int* FrameGrabber::getRgbValues() {
	return rgbValues;
}

bool* FrameGrabber::getValidValues() {
	return validValues;
}

int FrameGrabber::getValidCount() {
	return validCountTotal;
}
