/* This file is part of the Kiretu, a Kinect reconstruction tutor.
 * 
 * http://pille.iwr.uni-heidelberg.de/~kinect01/
 * 
 * The code ist licensed to you under the terms of the GNU General Public 
 * License, version 2.0. See the GPL2 file for the text of the license or the 
 * following URL:
 * 
 * http://www.gnu.org/licenses/gpl-2.0.txt
 * 
 * If you redistribute this file in source form, modified or unmodified, you
 * have to leave this header intact and distribute it under the same terms with 
 * the GPL2 file.
 * 
 * Binary distributions must follow the binary distribution requirements of
 * either License.
 * 
 */
 
 /** \file Hud.h
 * 
 * Header file of the HUD. The HUD (Head Up Display) is used for displaying 
 * various information to the application using OpenGL.
 * 
 * \author		Daniel Wunderlich (d.wunderlich@stud.uni-heidelberg.de)
 * \version		0.6
 * \date		2011-01-16
 */
 
#include <cstring>
#include <string>

#include <GL/glut.h>

#include "Util.h"


/**	\class	Hud 
 * 	\brief	C++-class which represents a HUD. The HUD (Head Up Display) is used 
 * 			for displaying various information to the application using OpenGL.
 * 	
 * 	\author	Daniel Wunderlich (d.wunderlich@stud.uni-heidelberg.de)
 * 	\version 0.6
 * 	\date	2011-01-16
 */
class Hud {
	
	public:
	
/**	\brief Constructor of the Hud-class.
 */ 
		Hud();

/**	\brief Deconstructor of the Hud-class.
 */ 
		~Hud();

/**	\brief 	Prints text using OpenGL's 
 * 			glutBitmapCharacter(void *font, int character) without 
 * 			specifying a certain position.
 *
 * 	\param 	text		The text which should be printed.
 * 	
 * 	\param 	printLarge	False: Text is printed with OpenGL's 
 * 						GLUT_BITMAP_HELVETICA_12. True: Text is printed with 
 * 						GLUT_BITMAP_HELVETICA_18.
 */ 		
		void printGLText(std::string text, bool printLarge);


/**	\brief 	Calls printGLText(std::string text, bool printLarge) after setting 
 * 			a specified position using OpenGL's 
 * 			glRasterPos2i(GLint x, GLint y).
 *
 * 	\param 	x			X-coordinate of the wanted text-position.
 * 	
 * 	\param 	y			Y-coordinate of the wanted text-position.
 * 	
 * 	\param 	text		The text which should be printed.
 * 	
 * 	\param 	printLarge	False: Text is printed with OpenGL's 
 * 						GLUT_BITMAP_HELVETICA_12. True: Text is printed with 
 * 						GLUT_BITMAP_HELVETICA_18.
 */ 
		void printGLTextPos(int x, int y, std::string text, bool printLarge);

/**	\brief 	Prints the labels 'Depth Map' and 'Video Output' using 
 * 			printGLTextPos(int x, int y, std::string text, bool printLarge).
 */ 
		void printGLLabel();


/**	\brief 	Prints a 'Scanning'-label in the middle of the depth-map-frame while 
 * 			scanning.
 * 
 * 	\param 	print	True: Label is printed. False: Label is not printed.
 */ 
		void printGLScanning(bool print);


/**	\brief 	Prints the help-box in the lower left corner of the window.
 * 	
 * 	\param 	print	True: Label is printed. False: Label is not printed.
 */ 
		void printGLHelp(bool print);


/**	\brief 	Prints the state-box in the lower right corner of the window.
 * 	
 * 	\param 	print	True: Label is printed. False: Label is not printed.
 * 	
 * 	\param 	angle	The current Kinect's angle.
 * 
 * 	\param	format	String with the current format of the video stream. Should 
 * 					be 'RGB', 'YUV-RGB' or 'IR'. 
 * 	
 * 	\param 	states	Array with 5 element indicating the steps of the pointcloud 
 * 					reconstruction. The meaning of the steps are explained at 
 * 					KinectCloud::getReconstructionSteps().
 */ 
		void printGLState(bool print, int angle, std::string format, bool states[5]);
};
