/* This file is part of the Kiretu, a Kinect reconstruction tutor.
 * 
 * http://pille.iwr.uni-heidelberg.de/~kinect01/
 * 
 * The code ist licensed to you under the terms of the GNU General Public 
 * License, version 2.0. See the GPL2 file for the text of the license or the 
 * following URL:
 * 
 * http://www.gnu.org/licenses/gpl-2.0.txt
 * 
 * If you redistribute this file in source form, modified or unmodified, you
 * have to leave this header intact and distribute it under the same terms with 
 * the GPL2 file.
 * 
 * Binary distributions must follow the binary distribution requirements of
 * the License.
 * 
 */
 
 /** \file YMLParser.cpp
 * 
 * Implementation of the YMLParser-class. See header-file for description.
 * 
 * \author		Daniel Wunderlich (d.wunderlich@stud.uni-heidelberg.de)
 * \version		0.8
 * \date		2011-01-26
 */


#include "YMLParser.h"

YMLParser::YMLParser(const std::string filename) : 
	filename(filename),
	text(""),
	rgbCam(3, std::vector<float>(3)), 
	rgbDist(5), 
	depthCam(3, std::vector<float>(3)), 
	depthDist(5), 
	rot(3, std::vector<float>(3)),
	trans(3) {
}

YMLParser::~YMLParser() {
}

void YMLParser::readFile() {
	
	char* filenameChar;
	
	filenameChar = new char[filename.size() + 1];
	strcpy(filenameChar, filename.c_str());
	
	std::ifstream f;
	f.open(filenameChar, std::ios::in);
		
	std::string line;
	
	while (!f.eof()) {
		getline(f, line);
		
		// Remove whitespace
		for (int i = 0; i < line.size(); i++) {
			if (line[i] == ' ') {
				line.erase(i, 1);
				i--;
			}
		}
		
		text += line;
	}
	
	f.close();
	delete[] filenameChar;
}

void YMLParser::findParam(KinectParam param) {
	
	std::string bracketl = "[";
	std::string bracketr = "]";
	std::string substring;
	std::string result = text;
	std::string rest;
	std::string::size_type pos;
	
	// Identify search string
	switch (param) {
		
		case RGB_INTRINSICS:
			substring = "rgb_intrinsics";
			break;
		case RGB_DISTORTION:
			substring = "rgb_distortion";
			break;
		case DEPTH_INTRINSICS:
			substring = "depth_intrinsics";
			break;
		case DEPTH_DISTORTION:
			substring = "depth_distortion";
			break;
		case ROT:
			substring = "R";
			break;
		case TRANS:
			substring = "T";
			break;
		default:
			substring = "";
			break;
	}
	
	std::cout << "Start: " << substring << std::endl;
	
	// Cut the text before param
	pos = result.find(substring);
	result = result.substr(pos);
	
	// Cut the text after right bracket
	pos = result.find("]");
	result = result.substr(0, pos);
	
	// Cut the text before left bracket
	pos = result.find(bracketl);
	result = result.substr(pos+1, result.length());
	
	// Split the comma separated values
	switch (param) {
		case RGB_INTRINSICS:
			csvToMatrix(result, rgbCam);
			break;
		case RGB_DISTORTION:
			csvToVector(result, rgbDist);	
			break;
		case DEPTH_INTRINSICS:
			csvToMatrix(result, depthCam);
			break;
		case DEPTH_DISTORTION:
			csvToVector(result, depthDist);	
			break;
		case ROT:
			csvToMatrix(result, rot);
			break;
		case TRANS:
			csvToVector(result, trans);	
			break;
		default:
			substring = "";
			break;
	}
}

void YMLParser::parseFile() {
	
	readFile();
	
	for (int i = 0; i < PARAM_END; i++) {
		findParam((KinectParam) i);
	}
	
}

void YMLParser::csvToVector(std::string str, std::vector<float> &vec) {
	
	std::cout << "size = " << vec.size() << std::endl;
	
	std::string value;
	std::string::size_type pos;
	int count = 0;
	
	str += ",";				// Append ',' for the last while-loop
	pos = str.find(",");	// For starting the while-loop
	
	// Loop over the string, as long as a comma is found
	while (pos != std::string::npos) {
		
		// Get the most left value
		value = str.substr(0, pos);
		
		// Save the value
		vec[count] = strToFloat(value);
				
		// Keep the rest of the string for the next loop
		str = str.substr(pos+1, str.length());
		
		// Get the comma-position for the next loop
		pos = str.find(",");	
		
		count++;
	}
}

void YMLParser::csvToMatrix(std::string str, std::vector< std::vector<float> > &mat) {
	
	std::vector<float> tempVec(9);
	int count = 0;
	
	csvToVector(str, tempVec);
		
	for (int i = 0; i < (int) mat.size(); i++) {
		for (int j = 0; j < (int) mat[i].size(); j++) {
			mat[i][j] = tempVec[count];
			count++;
		}
	}
}


std::vector< std::vector<float> > YMLParser::getRgbCam() {
	return rgbCam;
}

std::vector<float> YMLParser::getRgbDist() {
	return rgbDist;
}

std::vector< std::vector<float> > YMLParser::getDepthCam() {
	return depthCam;
}

std::vector<float> YMLParser::getDepthDist() {
	return depthDist;
}

std::vector< std::vector<float> > YMLParser::getRot() {
	return rot;
}

std::vector<float> YMLParser::getTrans() {
	return trans;
}
